/*
 * Goldwater Distributed Web Architecture.
 *
 * User Process Control
 *
 * Copyright 1996-2003 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:11 $
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>

#include <gwlog.h>
#include <gwreadconf.h>
#include <regstrings.h>

#include <gwusers.h>


/* NOTE: No logging is done in these functions as we cannot know that the
 *       logs have been set up at this point.  The usual place for user id
 *       manipulation is during initialisation.....
 */

/* ------------------------------------------------------------------------- */
/* Validate that a user and/or group exist
 * If uid is not NULL and uname is supplied, then uid is returned.
 * If gid is not NULL and group is supplied, then gid is returned.
 */
int
__gw_lookup_ugid( const char *uname, const char *group, uid_t *uid, gid_t *gid )
{
  struct passwd *pwd = NULL;
  struct group  *grp = NULL;

  if( uname && !(pwd = getpwnam(uname)) ) {gwerror = GWE_BADUSER;  return -1;}
  if( group && !(grp = getgrnam(group)) ) {gwerror = GWE_BADGROUP; return -1;}

  if( uid && pwd ) *uid = pwd->pw_uid;
  if( gid && grp ) *gid = grp->gr_gid;

  gwerror = GWE_NOERROR;
  return 0;
}

/* ------------------------------------------------------------------------- */

void
__gw_get_my_ugid( uid_t *uid, gid_t *gid )
{
  *uid = getuid();
  *gid = getgid();
}

/* ------------------------------------------------------------------------- */
/* Become the effective user
 */
int
__gw_become_effective_user( uid_t uid, gid_t gid )
{
  /* if( !uid && !gid ) { gwerror = GWE_BADUSER; return -1; } */
  
  if( setegid(gid) < 0 ) { gwerror = GWE_SETGROUP; return -1; }
  if( seteuid(uid) < 0 ) { gwerror = GWE_SETUSER;  return -1; }
  
  gwerror = GWE_NOERROR;
  return 0;
}

/* ------------------------------------------------------------------------- */
/* Become the user, forever - we loose the real userid
 */
int
__gw_become_user( uid_t uid, gid_t gid )
{
  if( !uid && !gid ) { gwerror = GWE_BADUSER; return -1; }
    
  if( setgid(gid) < 0 ) { gwerror = GWE_SETGROUP; return -1; }
  if( setuid(uid) < 0 ) { gwerror = GWE_SETUSER;  return -1; }
  
  gwerror = GWE_NOERROR;
  return 0;
}

/* ------------------------------------------------------------------------- */
/* Get the username from the config for the given domain index. 
 * This is the config index, as returned by __gw_find_domain_idx_in_config()
 */
int
__gw_get_user_from_config( int domlev, char *u, int l )
{
  char *user = NULL;
  char  path[256] = {0};
  long  len    = 0;
		   
  sprintf( path, "%s%06d%s", R_DOMAIN_B, domlev, R_USER_L );
			   
  gwerror = GWE_BADCNFG;
  if( gw_config_read_char( path, &user, &len ) >= 0 ) {
    if( len >= l ) return -1; /* => cos we need room for \0 */

    strncpy( u, user, len );
	u[len] = '\0';
    gwerror = GWE_NOERROR;
	return 0;
  } 
  return -1;
}

/* ------------------------------------------------------------------------- */
/* Get the usergroup from the config for the given domain index. 
 * This is the config index, as returned by __gw_find_domain_idx_in_config()
 */
int
__gw_get_group_from_config( int domlev, char *g, int l )
{
  char *group = NULL;
  char  path[256] = {0};
  long  len    = 0;
		   
  sprintf( path, "%s%06d%s", R_DOMAIN_B, domlev, R_GROUP_L );
			   
  gwerror = GWE_BADCNFG;
  if( gw_config_read_char( path, &group, &len ) >= 0 ) {
    if( len >= l ) return -1; /* => cos we need room for \0 */

    strncpy( g, group, len );
	g[len] = '\0';
    gwerror = GWE_NOERROR;
	return 0;
  } 
  return -1;
}

/* ------------------------------------------------------------------------- */
/* Tries to become root again (or whatever the real uid is).
 * Returns the old effective uid and gid
 */
int 
__gw_become_real_user( uid_t *wu, gid_t *wg )
{
  /* Remeber the euid and egid 
   */
  if(wu) *wu = geteuid();
  if(wg) *wg = getegid();

  /* Become ruid and rgrp again (the superuser) */
  return __gw_become_effective_user( getuid(), getgid() );
}


/* ------------------------------------------------------------------------- */
/* end */

