/*
 * Network Communication Layer 
 *
 * Multiplexed Network Client/Server Core Engine
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * -----------------------------------------------------------------------
 * $Id: gwneng.c,v 1.3 2003/04/10 17:21:50 csmith Exp $  $Date: 2003/04/10 17:21:50 $
 * -----------------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <stdio.h>

#include <phlib/xl_types.h>
#include <phlib/xl_socket.h>
#include <phlib/xl_ipaddr.h>
#include <phlib/xl_logging.h>

#include "sessions.h"
#include "pnetcom.h" /* Private Stuff */

#include "gwnevent.h"
#include "pxlapi.h"  /* Private stuff */
#include "gwxlapi.h"
#include "xlssl.h"

#include "logging.h"

static int still_alive = 1;
static int nagle_off   = 0;

static struct timeval tmv;
CTime currenttime = 0;  /* CoreWide Current Time (in seconds) */

/* ------------------------------------------------------------------------- */
/* -- THESE ARE PICKED UP BY THE XLAPI.C CODE --
 * The Code is Kept Separate Functionaly, but _must_ have access here!
 */
/* xlapi_set_event() referenced.... 
 */
fd_set _EventBits; 
int    _SFD       = 0;

/* xlapi_register_connection() referenced.... 
 */
int    _SelLimit = 0;

/* ------------------------------------------------------------------------- */


static NetworkConnectionData NetworkConnection[ SES_MAX ];

/* Static Globals - Fiddled with in a few local functions */
static fd_set prdfds, pwrfds, pexfds, selfds;
static fd_set rdfds,  wrfds,  exfds;

fd_set _ClosedConBits;

static int table[130] = {0}; /* log2 table */

static NetService  *NSrv = NULL;
static int _NumServ      = 0;     /* Number of server entries in NSrv */
static int _MaxConn      = 0;
static int _NSrvEnd      = 0;
static int fd_limit      = 0;

/* Prototypes */
static __inline__ void process_action( Action, NetService * );
static void close_session();

/* ---------------------------------------------------------------------- */

static void
comm_error( int err, char *str, int doexit )
{
  if(doexit == 1) {
    ph_logf( 0, "SERIOUS: Comms Error: %s", str );
    ph_close_log();
    exit(1);
  }
}

/* ---------------------------------------------------------------------- */

static int
get_sockets( )
{
  int    numserv = 0;
  int    port    = 0;
  int    sock    = 0;
  int    stat    = 0;
  static char buff[80];

  /* We end when we hit a service with a port number <= 0
   * This signifies either the end of the 'Listening' services
   * and the start of the 'manual connection outgoing' services
   * or the end of the service list iteslf.
   * Either way, this function finishes when it hits a port <= 0!
   */
  while( (port = NSrv[numserv].port) > 0 && numserv < _MaxConn ) {
    
    if( (sock = xl_bind_port_IP( port, NSrv[numserv].ipaddr )) == MINUSONE ) {
      sprintf( buff, "Server: binding to %s, port '%d'", 
	       NSrv[numserv].ipaddr, port );

      comm_error( 0, buff, 1 );
      return MINUSONE;
    }
	
	if( nagle_off && xl_sock_disable_nagle( sock ) >= 0 ) {
      ph_logf( 8, "Disabling Nagle Algorithm on %d", sock );
	}

    ph_logf( 9, "Accepting connections on %s:%d", 
	     NSrv[numserv].ipaddr, port );

    /*
     * now listen
     */
    if( listen( sock, Backlog ) < 0 ) {
      comm_error( errno, "Listen failure", 0 );
      return MINUSONE;
    }

    /* Make socket non-blocking */
    stat = fcntl( sock, F_GETFL, 0 );
    fcntl( sock, F_SETFL, stat | O_NONBLOCK );

    NSrv[numserv++].socket = sock; /* Store socket away with port num */
  }

  if( numserv >= _MaxConn ) {
    /* We can never blow the boundary of NSrv because it can never contain
	 * more elements than it can hold... obviously ...
	 * This message _will_ appear if we're full though, which is why I've
	 * left it in..... :o)
	 */
    ph_logf( 0, "Maximum number of connections reached [%d]", _MaxConn );
  }

  return numserv; /* Return number of network services..... */
}

/* ---------------------------------------------------------------------- */

RetValue
gwne_netcom_init( NetService *nsrv, int count, int max_conn, int wr_blk_size )
{
  /* Initialise log2 table
   */
  table[1  ] = 0;
  table[2  ] = 1;
  table[4  ] = 2;
  table[8  ] = 3;
  table[16 ] = 4;
  table[32 ] = 5;
  table[64 ] = 6;
  table[128] = 7;

  api_write_init( wr_blk_size ); /* Initialise write buffers */
  api_event_init(); /* Initialise event buffers */

  _MaxConn = max_conn;

  /* Initialise defered connection api
   * The data pointed to by the function arguments must remain intact!
   */
  api_defcon_init( nsrv, NetworkConnection );

  FD_ZERO( &prdfds );
  FD_ZERO( &pwrfds );
  FD_ZERO( &pexfds );
  FD_ZERO( &selfds );

  FD_ZERO( &_EventBits );

  _SelLimit = 0;

  NSrv    = nsrv;
  _NSrvEnd = count;

  return Success;
}

/* ---------------------------------------------------------------------- */

RetValue
gwne_netcom_deinit()
{
  api_event_deinit();
  return Success;
}

/* ---------------------------------------------------------------------- */

RetValue
gwne_netcom_init_server( int na_off )
{
  int ss      = 0;

  nagle_off = (na_off) ? 1 : 0;

  if( NSrv == NULL ) return Failure;

  /* Bind and listen on all ports required - Fills in NSrv */
  _NumServ = get_sockets(); /* Should pass in IP address too (VS) */

  if( _NumServ == MINUSONE ) return Failure;

  for( ss = 0; ss < _NumServ; ss++ ) {
    FD_SET( NSrv[ss].socket, &selfds );     /* Add 'listen' sockets to lst */
    if( NSrv[ss].socket >= _SelLimit ) _SelLimit = NSrv[ss].socket + 1; 
  }

  return Success;
}

/* ---------------------------------------------------------------------- */
/* DO NOT CALL THIS AFTER INIT_SERVER OR COM_POLL!!!!
 */
RetValue
gwne_netcom_add_handler( NetService *ns )
{
  int ix = 0;

  if( _NSrvEnd >= _MaxConn ) {
    ph_logf( 0, "Maximum number of connections reached [%d]", _MaxConn );
	return Failure;
  }

  if( ns->port <= 0 ) {
    /* If we're a defered connection, then add ourselves to the end of the
	 * NSrv list
	 */
    ns->port = 0;
	ix = _NSrvEnd;
  } else {
    /* Otherwise we're a 'server' and need to be with all the other server
	 * entries.
	 */
    for( ix = 0; ix < _NSrvEnd && NSrv[ix].port > 0; ix++ );
	/* Either the end of the list, or the end of the server section and the
	 * first of the defered connection section.
	 */
	NSrv[_NSrvEnd] = NSrv[ix]; /* Move the 'current' entry to the end */
  }

  NSrv[ix].port          = ns->port;
  NSrv[ix].ipaddr        = ns->ipaddr;
  NSrv[ix].desc          = ns->desc;
  NSrv[ix].svcidstr      = ns->svcidstr;
  NSrv[ix].read_handler  = ns->read_handler;
  NSrv[ix].event_handler = ns->event_handler;

  NSrv[++_NSrvEnd].port = NETSVCEND; /* Terminate list */

  return Success;
}

/* ---------------------------------------------------------------------- */
/* Blocktime of -1 forces 'always block when we don't have any events'
 *               0 forces 'block for the default of SEL_DELAY_SEC'
 *             +ve forces 'block for this many seconds'
 * Whilst there are any events pending blocktime is ignored and SEL_DELAY_SEC
 * is used instead
 */
void
gwne_netcom_poll( int loop_count, int block_time )
{
  register unsigned int bt, st, v, j;
  int                   ix, fd, ss;
  struct sockaddr_in    sin;
  int                   slen, ndfs;
  struct timeval        tv, *ptv;
  int                   gotevent;
  static char           ipbuf[129];
  EventMsg             *ev;
  Action                action;


  /* Reset our 'closed connection' fdset.
   * This is used purely to record what fds have dropped or have been
   * reassigned, so the calling application can act accordingly.
   */
  FD_ZERO(&_ClosedConBits);

  do {
    fd_limit = howmany( _SelLimit, NFDBITS ); /* No of fd bits (round to32 )*/

    /* ------ Treat events ------ */
    if( api_events_pending() > 0 ) {
    
      for( ix = 0; ix < fd_limit; ix++ ) { /* Treat blocks of 32bits */
	bt = ((int*)&_EventBits)[ix];       /* take 32bit block 'ix'  */
	while( bt ) {
	  j = bt & -bt;  /* Get LSb    */
	  bt -= j;       /* Remove LSb */
	  if     ( j < 0x100     ) v = table[j];
	  else if( j < 0x10000   ) v = table[j >> 8 ] + 8; 
	  else if( j < 0x1000000 ) v = table[j >> 16] + 16;
	  else                     v = table[j >> 24] + 24;
	  _SFD = v + (ix << 5);
    
	  maintain_event_queue( &ev, _SFD );
    
	  if( ev != NULL ) {
#ifdef DEBUG
	    printf( "  Calling event handler for session %d  (ptr: %x)\n", _SFD,
		    NSrv[ NetworkConnection[_SFD].ServiceIndex ].event_handler );
#endif
	    action = (*(NSrv[ NetworkConnection[_SFD].ServiceIndex ].event_handler))
	      ( ev, &(NetworkConnection[_SFD].PrivatePtr) );
      
#ifdef DEBUG
	    printf( "  Back from event handler for session %d\n", _SFD );
#endif
	    process_action( action, NSrv );
	  }
	}
      }
    } /* end of 'if events pending' */

    /* 
     * size is really rdlimit (blocks of 32) * bytes per block (fd_mask)
     * Since fd_mask is basically 4bytes, we can do a much faster
     * fd_limit <<2 instead of fd_limit * fd_mask(4)!
     */
    for( ix = 0; ix < fd_limit << 2; ix++ )
      ((unsigned int*)&rdfds)[ix] = ((unsigned int*)&prdfds)[ix] | 
	((unsigned int*)&selfds)[ix];
    
    /* memcpy( &rdfds, &prdfds, fd_limit <<2 ); */
    memcpy( &wrfds, &pwrfds, fd_limit <<2 );
    memcpy( &exfds, &pexfds, fd_limit <<2 );
    
    /* TODO: xlapi should remember the smallest time period assigned to an
     * event (across all sessions).  If we've got events pending, then we
     * should get select to block for this ammount of time.
     * This will mean that if we've only got timed events in the queue, instead
     * of spinning and calling maintain_event_queue repeatedly for events which
     * will not go-off for ages, select can block, quietly eating up the time
     * period and saving CPU clock cycles...
     */
    
    /* Set default select timeout - used when we have events pending */
    tv.tv_sec  = SEL_DELAY_SEC;
    tv.tv_usec = SEL_DELAY_USEC;
    
    gotevent = api_events_pending();

    /* If we've got events pending, then only allow select to block for the
	 * default time of SEL_DELAY_SEC.
	 * If we've been asked to block for 'block_time' seconds, then use this
	 * instead of blocking indefinately (events force the default blocktime
	 * to override the suggested blocktime as we want to process the events
	 * ASAP).
     */
    if( gotevent || (block_time >= 0) ) {
	  if( !gotevent ) tv.tv_sec = (block_time > 0) ? block_time : SEL_DELAY_SEC;
	  ptv = &tv;
    }
    else ptv = 0; /* No events pending or given blocktime, so sit on select  */

    /* The all important select */
    ndfs = select( _SelLimit, &rdfds, &wrfds, &exfds, ptv ); 

    if( ndfs < 0 ) comm_error( errno, "Select Failure", 0 );

    /* Update global second clock */
    gettimeofday( &tmv, 0 );
    currenttime = tmv.tv_sec;

    /*
     * If we timed out then loop 
     */
    if( ndfs == 0 ) continue;

    for( ix = 0; ix < fd_limit; ix++ ) { /* Treat blocks of 32bits */

      /* ------ Check RD fd's ------ */
      bt = ((int*)&rdfds)[ix];  /* take 32bit block 'ix' - Rd Pending FD's   */
      st = ((int*)&selfds)[ix]; /* - 'listen' pending FD's if we're a server */

      while( bt ) {
	j = bt & -bt;  /* Get LSb    */
	bt -= j;       /* Remove LSb */
	if     ( j < 0x100     ) v = table[j];
	else if( j < 0x10000   ) v = table[j >> 8 ] + 8; 
	else if( j < 0x1000000 ) v = table[j >> 16] + 16;
	else                     v = table[j >> 24] + 24;
	_SFD = v + (ix << 5);
	
	/* If we're a server, then we'll have one or more 'listen' fd's in our
	 * FD set.  Here we do a quick check to see if we've picked up one of
	 * these special fd's, and accept it if it is.
	 *
	 * This WILL NOT apply if this engine is being used in a client
	 * and will drop straight through.
	 */
	if( j & st ) { /* This fd is one of out 'listen' fd's so accept! */
	  /* 
	   * Ah Ha! A New Connection! 
	   */
      sin.sin_family = AF_INET;
	  slen = sizeof( sin );
	  fd = accept( _SFD, (struct sockaddr *)&sin, &slen );
	  
	  /*
	   * Fatal errors ???
	   */
	  if( fd < 0 ) {
	    if( errno != EINTR && errno != EWOULDBLOCK && errno != EAGAIN ) {
	      if( errno == EMFILE ) {
		    ph_logf(0,
		            "SERIOUS: Too many open files, rejecting accept");
	      } else comm_error( errno, "Accept Failure", 1 );
	    }
	  } else {       /* Valid ACCEPT completed */
	    
	    /* set fd check limit *//* What do we do when we close fd's?? */
	    
	    for( ss = 0; ss < _NumServ; ss++ ) {
	      /* look for connected socket in service list */
	      if( NSrv[ss].socket == _SFD ) { /* Find correct 'service' */
		/* Found correct service, so create an accepted fd to
		 * service index mapping */
		_SFD = fd;

		/* Create an index map from connection -> service, etc */
		NetworkConnection[ _SFD ].ServiceIndex = ss;
		NetworkConnection[ _SFD ].PrivatePtr   = 0;
    
		/* Are we creating an SSL connection???
		 */
#if 0
		if( NSrv[ss].ssl_ctx ) {
		  if( xlssl_begin_session( &NetworkConnection[_SFD], 
					 _SFD, NSrv[ss].ssl_ctx ) < 0 ) {
		    /* Error starting SSL session */
		  }
		}
#endif

		/* Setup an initial event - New Session */
		xlapi_register_event( EV_NEW_SESSION, 0 ); 
    
		ph_logf( 9, "## Connect from %s:%d ( %d ) ##", 
			 xl_get_IPAddr_str( ipbuf, 
					    (ulong)sin.sin_addr.s_addr ), 
			 NSrv[ss].port, fd); 

		if( _SFD >= _SelLimit ) _SelLimit = _SFD + 1; 
		break;
	      }
	    }
	    if( NSrv[ss].socket >= _SFD ) { /* This SHOULD NEVER happen */
	      ph_logf( 0, 
		       "EEK - Can't match to service (Accept on %d)", _SFD );
	      ph_logf( 0, 
		      "SERIOUS: "
		      "Cannot match to service (accept on %d)", _SFD );
	    }
	  }
	} else { /* Process read request */
	  /* Index into connection array for this service is given by:
	   * NetworkConnection[_SFD].ServiceIndex
	   */
	  
	  action = (*(NSrv[ NetworkConnection[_SFD].ServiceIndex ].read_handler))
	    ( _SFD, &(NetworkConnection[_SFD].PrivatePtr) );

	  process_action( action, NSrv );
	}
      }
      
      /* ------ Check WR fd's ------ */
      bt = ((int*)&wrfds)[ix]; /* take 32bit block 'ix' */
      while( bt ) {
	j = bt & -bt;  /* Get LSb    */
	bt -= j;       /* Remove LSb */
	if     ( j < 0x100     ) v = table[j];
	else if( j < 0x10000   ) v = table[j >> 8 ] + 8; 
	else if( j < 0x1000000 ) v = table[j >> 16] + 16;
	else                     v = table[j >> 24] + 24;
	_SFD = v + (ix << 5);

	if( maintain_write_buffer( _SFD ) < 0 ) {
	  process_action( FREE_SESSION, NSrv );    
	}
      }
    }
  } while( still_alive && (loop_count == 0 || --loop_count) );
}

/* ---------------------------------------------------------------------- */

static __inline__  void
process_action( Action action, NetService *NSrv )
{
  /* Assume _SFD is the fd of the current session */

  switch( action ) {

  case FREE_SESSION:
    close_session( ); /* Session is in _SFD */
    break;

  case TERMINATE_SESSION: /* All cleanup MUST have been done by now */
    ph_logf( 9, "## Lost Connection for session %d [%s] ##", _SFD,
	     NSrv[ NetworkConnection[_SFD].ServiceIndex ].svcidstr );

    /* Do these again, just in case... */
    FD_CLR( _SFD, &prdfds );     /* Remove from prdfds list    */
    FD_CLR( _SFD, &pwrfds );     /* Remove from pwrfds list    */
    FD_CLR( _SFD, &pexfds );     /* Remove from pexfds list    */
    FD_CLR( _SFD, &rdfds );      /* Remove from rdfds list     */
    FD_CLR( _SFD, &wrfds );      /* Remove from wrfds list     */
    FD_CLR( _SFD, &exfds );      /* Remove from exfds list     */
    FD_CLR( _SFD, &_EventBits ); /* Remove from evfds list     */

    NetworkConnection[_SFD].ServiceIndex = -1;/* Remove index from srvindex */
    NetworkConnection[_SFD].PrivatePtr   = 0; /* Clear wrokspace ptr        */

#if 0    
    xlssl_end_session( &(NetworkConnection[_SFD].SSLCon) );
#endif

#if 0 
    /* Close our end of the connection (FIN to client)
     */
    if( shutdown( _SFD, 1 ) == 0 ) {
      /* Okay close of 'send' channel succeeded... */
      /*
       * This is a great idea in practice....... but can't be done
       * like this because the server will block until it's closed
       * down the connection.
       * The trick will be to close down the send channel, mask off
       * the 'wr' flags in the FD array, register and event to do the
       * clean up of the read channel....  Ah, but the service handler
       * will still be receiving connections from the read channel.
       * Arg!
       * Nasty Nasty.
       * We need to 'hotwire' the read_handler for certain _SFD's so
       * that we enter the 'read to junk' bit prior to closing the
       * connection down completely.
       * In which case the event registered will be a timed event, which
       * if triggered *before* we've unbuffered the pending read then
       * we just read (if we can) and close the connection).
       * Else we read what is pending and remove the event.
       * Ah! Now, we could do this anyway:
       * 1. Disable all read/write's from the _SFD (FD_ZERO etc)
       * 2. Register timed event
       * 3. When event goes off, test _SFD for pending reads and
       *    pull them off.
       * 4. Close connection completely.
       *
       * The only problem is the size of the data that might be pending.
       * Do we want to read _everything_ until there is no more, and then
       * close??  This is terrible if we do!
       */
      fd_set cln_rd;
      struct timeval tv;
      char   junk[1024];

      FD_ZERO(&cln_rd);
      FD_SET(_SFD, &cln_rd);
      do {
        tv.tv_sec = 2;
        tv.tv_usec = 0;
      while ( select( _SFD+1, &cln_rd, NULL, NULL, &tv ) > 0 &&
             FD_ISSET(_SFD,&cln_rd) &&
             read( _SFD, junk, sizeof( junk ) ) > 0 );
      

    }
#endif
    close( _SFD );               /* Close fd                   */

    /* Record the fd that we've just closed in our special Closed Connection
	 * fd set.  This is used by the calling application to detect what fd's
	 * have been dropped (though they may have been reassigned already).
	 */
    FD_SET( _SFD, &_ClosedConBits );
    break;
    
  case ENABLE_WRITE:
    FD_SET( _SFD, &pwrfds );     /* Add write fd to list */
    break;

  case DONE_WRITE:
    FD_CLR( _SFD, &pwrfds );     /* Remove write fd from list */
    FD_CLR( _SFD, &wrfds );
    break;
    
  case DISABLE_READ:
    FD_CLR( _SFD, &prdfds );     /* Remove from prdfds list    */
    break;
    
  case ENABLE_READ:
    FD_SET( _SFD, &prdfds );     /* Add read fd to prdfds list    */
    break;
    
  case TERMINATE_SERVER:
    still_alive = 0;            /* End server */
    break;

  case FINISHED:
  case KEEP_CONNECTION:
    break;
    
  default:
    ph_logf( 0, "ERR: Cannot process Action type: %d", (int)action );
    break;
  }
}

/* ---------------------------------------------------------------------- */

void
_netcom_set_wrfd( int fd ) 
{
  FD_SET( fd, &pwrfds ); /* Add write fd to list */
  FD_SET( fd, &pexfds );
}

/* ---------------------------------------------------------------------- */

void
_netcom_reset_wrfd( int fd ) 
{
  FD_CLR( fd, &pwrfds ); /* Add write fd to list */
  FD_CLR( fd, &pexfds );
}

/* ---------------------------------------------------------------------- */

static void
close_session()
{
  FD_CLR( _SFD, &prdfds );     /* Remove from prdfds list    */
  FD_CLR( _SFD, &pwrfds );     /* Remove from pwrfds list    */
  
  FD_CLR( _SFD, &rdfds );      /* Remove from rdfds list     */
  FD_CLR( _SFD, &wrfds );      /* Remove from wrfds list     */
  
  FD_CLR( _SFD, &pexfds );     /* Remove exceptions */
  FD_CLR( _SFD, &exfds );
  
  FD_CLR( _SFD, &_EventBits );  /* Remove from evfds list     */
  
  flush_events( _SFD );
  flush_writes( _SFD );
  
  xlapi_register_event( EV_END_SESSION, 0 ); /* Force FREE_CONNECTION action*/
}

/* ---------------------------------------------------------------------- */

void
gwne_get_closed_fd( /*callback function - closed_fd_process( int fd ) */ )
{
  register unsigned int bt, v, j;
  int                   ix, fd;

  for( ix = 0; ix < fd_limit; ix++ ) { /* Treat blocks of 32bits */
	bt = ((int*)&_ClosedConBits)[ix];  /* take 32bit block 'ix'  */
	while( bt ) {
	  j = bt & -bt;  /* Get LSb    */
	  bt -= j;       /* Remove LSb */
	  if     ( j < 0x100     ) v = table[j];
	  else if( j < 0x10000   ) v = table[j >> 8 ] + 8; 
	  else if( j < 0x1000000 ) v = table[j >> 16] + 16;
	  else                     v = table[j >> 24] + 24;
	  fd = v + (ix << 5);

	  /* Process fd here - perform callback */
    }
  }
}

/* ---------------------------------------------------------------------- */
/* end */
